# Generated from src-upstream/CPYTHON_REVISION:
UPSTREAM_REVISION = 'a5a7f5e16d8c3938d266703ea8fba8ffee3e3ae5'
# Generated from src-upstream/CPYTHON_BRANCH:
UPSTREAM_BRANCH = main
# Generated from src-upstream/CPYTHON_VERSION:
UPSTREAM_VERSION = (3, 14)
